"""
main parski module
"""
from __future__ import print_function

import sys
import re

from filter_datetime_mod import filter_datetime
from path_dict_conv_mod import dict_to_paths

def filter_data(data, search_input, max_results=5000, metrics=False):
    '''
    filter data based on path list or search dict arrays
    '''

    ##check if single path or single path group was used
    try:
        search_input['path']
        search_input['value']
        print("single path")
        search_input = [search_input]
    except:
        pass

    try:
        search_input[0]['path']
        search_input[0]['value']
        print("single path group")
        search_input = [search_input]
    except:
        pass

    ##check if single dict was used
    if isinstance(search_input, dict):
        print("Input is dict")
        search_input = [search_input]

    print("Data length: ", len(data))
    print("Running Search...")
    filtered_results = []
    global metrics_dict
    metrics_dict = {'missing_keys': 0, 'regex_errors': 0}

    for i, search_list in enumerate(search_input):
        if isinstance(search_list, dict):
            search_list = dict_to_paths(search_list, [], [])
        elif isinstance(search_list, list):
            search_list = search_list
        else:
            print("Invalid input type.  Must be 'dicts' or 'lists'")
            sys.exit(1)
        print("search_list: ", search_list)
        for search in search_list:
            try:
                int(search['path'][0])
            except:
                search['path'] = [i] + search['path']

        #Remove paths where value is None
        temp_paths = []
        for path in search_list:
            if path['value'] != str(None):
                temp_paths.append(path)
        search_list = temp_paths
        for entry in data:
            if len(filtered_results) >= max_results:
                break
            elif __entry_filter(entry, search_list, True) and entry not in filtered_results:
                filtered_results.append(entry)
    print("Results found: ", len(filtered_results))
    if metrics:
        return {'results': filtered_results, 'metrics': metrics_dict}
    return filtered_results

def __entry_filter(entry, path_list, first):
    '''
    filter single entry with path_list
    '''

    path_groups = []
    #loop through all paths in path_list
    for path in path_list:
        #check if path has ended.  If so, remove from checks if passed, return False if failed
        if len(path['path']) <= 1:
            #One fail is total fail.  One pass is NOT total pass.
            if __value_check(entry, path) is False:
                return False

        #sort remaining paths into groups based on first value
        else:
            found = False
            for group in path_groups:
                if path['path'][0] == group[0]['path'][0]:
                    group.append(path)
                    found = True
                    break
            if not found:
                path_groups.append([path])

    #run each group together
    return __filter_path_groups(entry, path_groups, first)

def __value_check(entry, path):
    '''
    check if end value is equal to the the one defined in the path
    '''

    global metrics_dict
    value = str(path['value'])
    if value == 'null':
        value = 'None'
    entry_index = path['path'][0]

    ##check for key errors
    try:
        entry[entry_index]
    except:
        metrics_dict['missing_keys'] += 1
        return False

    ##check regex and for compile errors
    try:
        regex_value = re.compile(value, re.IGNORECASE)
        if not re.search(regex_value, str(entry[entry_index]).encode('utf-8')) and not filter_datetime(value, entry[entry_index]):
            return False
        return True
    except:
        metrics_dict['regex_errors'] += 1
        return False

def __filter_path_groups(entry, path_groups, first):
    '''
    Groups checks by list to ensure passes are at same index
    '''
    if entry is None:
        return False
    elif not first:
        for group in path_groups:
            current_index = group[0]['path'][0]

            #create new groups with first path removed
            for i, path in enumerate(group):
                group[i] = {'value': path['value'], 'path': path['path'][1:]}

            #check if list, if so run against all
            if isinstance(entry, list):
                match = False
                for item in entry:
                    if __entry_filter(item, group, False):
                        match = True
                if not match:
                    return False
            else:
                if current_index in list(entry):
                    if not __entry_filter(entry[current_index], group, False):
                        return False
                else:
                    return False
        return True
    else:
        for group in path_groups:
            for i, path in enumerate(group):
                group[i] = {'value': path['value'], 'path': path['path'][1:]}
            if __entry_filter(entry, group, False):
                return True



from __future__ import print_function
from os import environ
from time import sleep
import json
import requests


API_RETRY = 3

def get_data(url=None, file_name='output.json', source="file", key_var="PCT_API_READ_KEY"):
    '''
    Load data from either a source file or the api
    '''
    if source == "file":
        try:
            print("Loading File...")
            with open(file_name, 'rb') as stored_file:
                data = stored_file.read()
                return {'status_code': 200, 'error_msg': None, 'data': json.loads(data)}
        except Exception as error:
            print("File could not be loaded.  Pulling from api")
            if url is None:
                return {'status_code': 404, 'error_msg': 'File could not be loaded', 'data': None}
            return _api_load(url, file_name)
    elif source == "api":
        return _api_load(url, file_name, key_var)
    else:
        return {'status_code': 500, 'error_msg': "Invalid source parameter.", 'data': None}

def _write_data(data, file_name='output.json'):
    with open(file_name, 'wb') as output_file:
        output_file.write(json.dumps(data, sort_keys=True, indent=4))
        output_file.close()
        print("File written successfully...")

def _api_load(url, file_name='output.json', key_var="PCT_API_READ_KEY"):
    return_request = {}
    return_request['data'] = []
    return_request['error_msg'] = None
    return_request['status_code'] = 200
    try:
        API_KEY = environ[key_var]
    except KeyError:
        return_request['status_code'] = 401
        return_request['error_msg'] = key_var + ' could not be loaded.  Unauthorized.'
        return_request['data'] = None
        return return_request
    headers = {
        'x-api-key': API_KEY,
        'Cache-Control': 'no-cache'
    }
    response, resp_headers = _api_try(url, headers, None)
    if response['data'] is None:
        return response

    # Get page count, output if none
    try:
        page_count = int(resp_headers.get("x-pagination-page-count"))
    except:
        print("No page count specified.  Writing Single Page...")
        json_string = response['data']
        if json_string is None:
            return response
        else:
            clean_json_string = '[' + json_string.lstrip('[').rstrip(']') + ']'
            _write_data(json.loads(clean_json_string), file_name=file_name)

    print("%s pages to be downloaded" % page_count)
    for page in range(1, page_count+1):
        response, response_headers = _api_try(url, headers, page)
        if response['data']:
            json_string = response['data']
            clean_json_string = '[' + json_string.lstrip('[').rstrip(']') + ']'
            return_request['data'] = return_request['data'] + json.loads(clean_json_string)
        else:
            return response

    print(page_count)
    _write_data(return_request['data'], file_name=file_name)
    return return_request


def _api_try(url, headers, page=None):
    tries = 0
    if page:
        print("\nPage: %s" % page)
    while tries < API_RETRY:
        print("Sending API request")
        response, resp_headers = _api_call(url, headers, page)
        if response['error_msg']:
            print(response['error_msg'])
        if response['status_code'] == 401 or response['status_code'] == 404:
            return response, None
        if response['data']:
            print("Returned Successfully")
            return response, resp_headers
        print("Trying again in %s seconds\n" % ((tries+1)**2))
        sleep((tries+1)**2)
        print("LET'S DO THIS!!!!")
        tries += 1

    response['error_msg'] = "Amount of Fails Exceeded.  The API just doesn't want to talk to you today."
    return response, resp_headers


def _api_call(url, headers, page=None):
    return_request = {}
    return_request['data'] = None
    return_request['error_msg'] = None

    if page is not None:
        url += "?Page=" + str(page)
    try:
        req = requests.get(url, headers=headers)
    except:
        return_request['error_msg'] = 'API Server could not be reached'
        return_request['status_code'] = 404
        return return_request, None
    if req.status_code == 404:
        return_request['error_msg'] = 'API Server could not be reached'
        return_request['status_code'] = 404
        return return_request, None

    return_request['status_code'] = req.status_code
    if req.status_code != 200:
        print("WARNING:  API call status code was %s" % req.status_code)

    if req.status_code == 401:
        return_request['error_msg'] = "Unauthorized.  Your API key isn't working..."
    elif req.status_code == 404:
        return_request['error_msg'] = "Not found.  Your url is probably wrong or you're not on VPN"
    elif req.status_code == 502:
        return_request['error_msg'] = "The response failed.  This happens from time to time..."
    elif req.status_code == 503:
        return_request['error_msg'] = "Internal Server Error.  The API might be down..."
    else:
        try:
            return_request['data'] = json.dumps(req.json())
        except:
            return_request['error_msg'] = 'Returned request could not be converted to json'

    return return_request, req.headers

